#  Copyright 2022 VMware, Inc.
#  SPDX-License-Identifier: Apache-2.0
import asyncio
import sys
from typing import Union

import aria.ops.adapter_logging as logging
from alerts import add_alerts
from aoa_client import AOAClient
from aoa_client import format_url
from aoa_client import RestRequestException
from aria.ops.adapter_instance import AdapterInstance
from aria.ops.definition.adapter_definition import AdapterDefinition
from aria.ops.result import CollectResult
from aria.ops.result import EndpointResult
from aria.ops.result import RelationshipUpdateModes
from aria.ops.result import TestResult
from aria.ops.timer import Timer
from cluster import Cluster
from cluster import get_clusters
from constants import ADAPTER_KIND
from constants import ADAPTER_NAME
from constants import ARIA_APPS_CREDENTIAL_KEY
from constants import ARIA_APPS_TOKEN_KEY
from constants import ARIA_APPS_URL
from constants import ARIA_OPS_URL
from constants import CSP_CREDENTIAL_KEY
from constants import CSP_TOKEN_KEY
from constants import CUSTOM_METRIC_PREFIX
from constants import DATA_RETENTION_TIME
from constants import ENABLE_EXTERNAL_LINKS
from constants import ENABLE_POD_COLLECTION
from constants import INVALID_EXTERNAL_LINK_ID
from constants import PLATFORM_DEFAULT
from constants import QUERY_TIMEOUT
from deployment import add_deployment_relationships
from deployment import Deployment
from deployment import get_deployments
from ec2_instance import get_ec2_instances
from external_links import sync_external_links
from httpx import ConnectError
from httpx import ConnectTimeout
from namespace import add_namespace_relationships
from namespace import get_namespaces
from namespace import Namespace
from node import add_node_relationships
from node import get_nodes
from node import Node
from pod import add_pod_relationships
from pod import delete_not_existing_pods
from pod import get_pods
from pod import Pod
from replica_set import add_replicaset_relationships
from replica_set import get_replicasets
from replica_set import ReplicaSet
from virtual_machine import get_virtual_machines

logger = logging.getLogger(__name__)


async def test(adapter_instance: AdapterInstance) -> TestResult:
    logger.info("Starting test")
    result = TestResult()
    custom_metric_prefixes = ["kubernetes"]
    enable_external_links = (
        adapter_instance.get_identifier_value(ENABLE_EXTERNAL_LINKS, "True") == "True"
    )
    if adapter_instance.get_identifier_value(CUSTOM_METRIC_PREFIX):
        custom_metric_prefixes = adapter_instance.get_identifier_value(
            CUSTOM_METRIC_PREFIX
        ).split(",")
    try:
        async with AOAClient(adapter_instance) as aoa_client:
            # query test
            for custom_metric_prefix in custom_metric_prefixes:
                custom_metric_prefix = custom_metric_prefix.strip()
                if len(custom_metric_prefix) < 1:
                    logger.info("Skipping empty prefix")
                    continue
                logger.info(
                    f"Testing with custom metric prefix '{custom_metric_prefix}'"
                )
                response = await aoa_client.run_query(
                    f"""
                    lag(10m, count(align(60s, count(
                        ts('{custom_metric_prefix}.deployment.ready_replicas'),
                        cluster,
                        namespace_name,
                        deployment
                    ))))
                    """,
                )
                logger.debug("Received test connection query response:")
                logger.debug(response)
                if "timeseries" not in response:
                    result.with_error(
                        f"No timeseries found for '{custom_metric_prefix}'. Verify PROMQL query '{custom_metric_prefix}.deployment.ready_replicas' returns data in VMware Aria Operations for Applications"
                    )
                    # Integration SDK will only return one error at a time. If we have an error, we should just return it immediately.
                    return result
                logger.info(
                    f"Finished test for prefix '{custom_metric_prefix}' successfully"
                )
                logger.debug(
                    f"Test result for custom metric prefix '{custom_metric_prefix}': {result.get_json()}"
                )
            # external link test
            if enable_external_links:
                try:
                    logger.warning(
                        "Testing 'External Link' permissions. The following 'delete' call "
                        "is expected to fail with a 404 error, which indicates sufficient "
                        "permissions."
                    )
                    await aoa_client.delete(
                        f"/api/v2/extlink/{INVALID_EXTERNAL_LINK_ID}"
                    )
                except RestRequestException as e:
                    # This will return a 404 error if the token has permission to modify
                    # external links. Otherwise, it will return a 403 error.
                    if e.status_code == 404:
                        pass
                    else:
                        raise e
                    logger.debug("Received test connection query response:")
                    logger.debug(response)
            else:
                logger.info("Skipping external link test")
    except Exception as e:
        handle_aoa_client_exceptions(e, result, enable_external_links)
    logger.info("Finished test")
    logger.debug(f"Returning test result {result.get_json()}")
    return result


def handle_aoa_client_exceptions(
    e: Exception, result: Union[TestResult, CollectResult], enable_external_links
) -> Union[TestResult, CollectResult]:
    # Unknown host (no DNS entry)
    if type(e) is ConnectError and "Name or service not known" in repr(e):
        logger.error("Connect Error: Name or service not known")
        result.with_error(
            "Unknown VMware Aria Operations for Applications URL. Make sure the hostname is correct, "
            "has a DNS entry, and is reachable from the Aria Operations Cloud Proxy."
        )

    # Known host that is off or moved (DNS entry is out of date)
    # Wavefront env with valid token that takes too long to respond
    elif type(e) is ConnectTimeout:
        logger.error("Connection Timed out")
        # TODO: Implement timeout parameter in advanced config and adjust message
        #  https://jira.eng.vmware.com/browse/VOPERATION-39470
        result.with_error(
            "Connection to VMware Aria Operations for Applications URL timed out. Make sure the host "
            "exists and is reachable from the Aria Operations Cloud Proxy. If the URL "
            "is valid, the API may be taking too long to respond."
        )

    # Known host that is not a wavefront environment (closed 443 port)
    elif type(e) is ConnectError and "All connection attempts failed" in repr(e):
        logger.error("Connect Error: All connection attempts failed")
        result.with_error(
            "Connection to VMware Aria Operations for Applications URL could not be established on "
            "port 443. Make sure the host is VMware Aria Operations for Applications."
        )

    # Known host that is not a wavefront environment (open 443 port)
    elif type(e) is RestRequestException and e.status_code == 404:
        logger.error("Connection returned 404")
        result.with_error(
            "Connection to VMware Aria Operations for Applications URL does not have expected "
            "endpoint(s). Make sure the host is VMware Aria Operations for "
            "Applications."
        )

    # Wavefront env with bad token
    elif type(e) is RestRequestException and e.status_code == 401:
        logger.error("Connection returned 401")
        result.with_error(
            "VMware Aria Operations for Applications authorization token is invalid."
        )

    # Wavefront env with token with insufficient permissions
    elif type(e) is RestRequestException and e.status_code == 403:
        logger.error("Connection returned 403")
        if enable_external_links:
            message = (
                "VMware Aria Operations for Applications authorization token has insufficient "
                "permissions. Make sure the token has the 'Metric' and 'External "
                "Links' permissions."
            )
        else:
            message = (
                "VMware Aria Operations for Applications authorization token has insufficient "
                "permissions. Make sure the token has the 'Metric' permission."
            )
        result.with_error(message)

    else:
        error_type = "connection test" if type(result) is TestResult else "collection"
        logger.error(f"Unexpected {error_type} error")
        logger.exception(e)
        result.with_error(f"Unexpected {error_type} error: " + repr(e))
    return result


def get_endpoints(adapter_instance: AdapterInstance) -> EndpointResult:
    logger.info("Starting getEndpoints")
    result = EndpointResult()
    # In the case that an SSL Certificate is needed to communicate to the target,
    # add each URL that the adapter uses here. Often this will be derived from a 'host'
    # parameter in the adapter instance. In this Adapter we don't use any HTTPS
    # connections, so we won't add any. If we did, we might do something like this:
    # result.with_endpoint(adapter_instance.get_identifier_value("host"))
    #
    # Multiple endpoints can be returned, like this:
    # result.with_endpoint(adapter_instance.get_identifier_value("primary_host"))
    # result.with_endpoint(adapter_instance.get_identifier_value("secondary_host"))
    #
    # This 'get_endpoints' method will be run before the 'test' method,
    # and VMware Aria Operations will use the results to extract a certificate from
    # each URL. If the certificate is not trusted by the VMware Aria Operations Trust
    # Store, the user will be prompted to either accept or reject the certificate. If
    # it is accepted, the certificate will be added to the AdapterInstance object that
    # is passed to the 'test' and 'collect' methods. Any certificate that is
    # encountered in those methods should then be validated against the certificate(s)
    # in the AdapterInstance.
    result.with_endpoint(
        format_url(adapter_instance.get_identifier_value(ARIA_APPS_URL))
    )
    logger.info("Finished getEndpoints")
    logger.debug(f"Returning endpoints {result.get_json()}")
    return result


async def collect(adapter_instance: AdapterInstance) -> CollectResult:
    result = CollectResult()
    collect_pods = (
        adapter_instance.get_identifier_value(ENABLE_POD_COLLECTION, "True") == "True"
    )
    enable_external_links = (
        adapter_instance.get_identifier_value(ENABLE_EXTERNAL_LINKS, "True") == "True"
    )
    try:
        result.update_relationships = RelationshipUpdateModes.PER_OBJECT
        custom_metric_prefixes = ["kubernetes"]
        if adapter_instance.get_identifier_value(CUSTOM_METRIC_PREFIX):
            custom_metric_prefixes = adapter_instance.get_identifier_value(
                CUSTOM_METRIC_PREFIX
            ).split(",")
            custom_metric_prefixes = [
                prefix.strip() for prefix in custom_metric_prefixes
            ]

        logger.info("Starting collection")
        async with AOAClient(adapter_instance) as aoa_client:
            logger.info(
                "Collecting objects from VMware Aria Operations for Applications"
            )
            with Timer(logger, "Adding clusters"):
                clusters = await get_clusters(aoa_client, custom_metric_prefixes)
                result.add_objects(clusters.values())

            with Timer(logger, "Adding nodes"):
                nodes = await get_nodes(aoa_client, custom_metric_prefixes)
                result.add_objects(nodes.values())

            with Timer(logger, "Adding namespaces"):
                namespaces = await get_namespaces(aoa_client, custom_metric_prefixes)
                result.add_objects(namespaces.values())

            with Timer(logger, "Adding deployments"):
                deployments = await get_deployments(aoa_client, custom_metric_prefixes)
                result.add_objects(deployments.values())

            with Timer(logger, "Adding replicasets"):
                replicasets = await get_replicasets(aoa_client, custom_metric_prefixes)
                result.add_objects(replicasets.values())

            if collect_pods:
                with Timer(logger, "Adding pods"):
                    pods = await get_pods(aoa_client, custom_metric_prefixes)
                    result.add_objects(pods.values())
                    with Timer(logger, "Adding pod relationships"):
                        add_pod_relationships(pods, namespaces, nodes, replicasets)
            else:
                logger.info("Skipping pod collection")

            # Add alerts only after creating all objects to ensure alerts can
            # associate with their parent object
            with Timer(logger, "Adding alerts"):
                await add_alerts(aoa_client, result, collect_pods)

            with adapter_instance.suite_api_client as suite_api_client:
                if enable_external_links:
                    with Timer(logger, "Syncing external links"):
                        aria_apps_url = format_url(
                            adapter_instance.get_identifier_value(ARIA_APPS_URL)
                        )
                        aria_ops_url = format_url(
                            adapter_instance.get_identifier_value(ARIA_OPS_URL)
                        )
                        await sync_external_links(
                            aoa_client, suite_api_client, aria_apps_url, aria_ops_url
                        )
                else:
                    logger.info("Skipping external links")

                # Relationships are not collected on the same collection as new objects
                # are added, so we'll collect on the second collection (1 instead of 0)
                # so that we don't need to wait ~1 hour to get relationships initially.
                compute_relationships = adapter_instance.collection_number % 10 == 1
                if compute_relationships:
                    result.update_relationships = RelationshipUpdateModes.ALL
                    with Timer(logger, "Adding node relationships"):
                        add_node_relationships(nodes, clusters)
                    with Timer(logger, "Adding namespace relationships"):
                        add_namespace_relationships(namespaces, clusters)
                    with Timer(logger, "Adding deployment relationships"):
                        add_deployment_relationships(deployments, namespaces)
                    with Timer(logger, "Adding replicaset relationships"):
                        add_replicaset_relationships(replicasets, deployments)

                    with Timer(logger, "Computing external relationships"):
                        # Get Virtual Machines related to the given node_ips
                        with Timer(logger, "Adding VM relationships"):
                            result.add_objects(
                                get_virtual_machines(suite_api_client, nodes)
                            )
                        # Get EC2 Instances related to given node provider_ids
                        with Timer(logger, "Adding EC2 instance relationships"):
                            result.add_objects(
                                get_ec2_instances(suite_api_client, nodes)
                            )

                    # Cleaning up pods is not computing relationships, but we may as
                    # well do it on the same schedule.
                    if collect_pods:
                        with Timer(logger, "Cleaning up old pods"):
                            delete_not_existing_pods(suite_api_client, adapter_instance)
                    else:
                        logger.info("Skipping pod cleanup")

    except Exception as e:
        handle_aoa_client_exceptions(e, result, enable_external_links)
    logger.info("Finished collection")
    logger.debug(f"Returning collection result {result}")
    return result


def get_adapter_definition() -> AdapterDefinition:
    """
    The adapter definition defines the object types and attribute types
    (metric/property) that are present in a collection. Setting these object types and
    attribute types helps VMware Aria Operations to validate, process, and display the
    data correctly.
    :return: AdapterDefinition
    """
    logger.info("Starting getAdapterDefinition")
    definition = AdapterDefinition(ADAPTER_KIND, ADAPTER_NAME)

    apps_credential = definition.define_credential_type(
        ARIA_APPS_CREDENTIAL_KEY, "VMware Aria Operations for Applications Credential"
    )
    apps_credential.define_password_parameter(
        ARIA_APPS_TOKEN_KEY, "VMware Aria Operations for Applications Token", True
    )

    # CSP
    csp_credential = definition.define_credential_type(
        CSP_CREDENTIAL_KEY, "VMware CSP User Token"
    )
    csp_credential.define_password_parameter(
        CSP_TOKEN_KEY, "VMware Cloud Services Token", True
    )

    definition.define_string_parameter(
        ARIA_APPS_URL,
        label="VMware Aria Operations for Applications URL",
        description="URL of VMware Aria Operations for Applications instance.",
        required=True,
    )
    definition.define_string_parameter(
        ARIA_OPS_URL,
        label="VMware Aria Operations URL",
        description="URL of VMware Aria Operations instance. Everything before '/ui/'.",
        required=True,
        advanced=True,
        default="https://www.mgmt.cloud.vmware.com/vrops-cloud",
    )

    definition.define_string_parameter(
        CUSTOM_METRIC_PREFIX,
        label="Custom Metric Prefix(es)",
        description="Comma-separated list of custom metric prefixes. The Kubernetes Integration allows you to define a custom metric prefix. If not set, the default is 'kubernetes'. Different Kubernetes clusters may be configured with different custom metric prefixes.",
        required=False,
        advanced=True,
        default="kubernetes",
    )

    definition.define_enum_parameter(
        ENABLE_POD_COLLECTION,
        values=["True", "False"],
        label="Enable Pod Collection",
        description="Determines whether to collect Pods from VMware Aria Operations for Applications. If set to False, pod objects and Namespace pod metrics will not be collected, and pod alerts will be assigned to a parent object.",
        required=True,
        advanced=True,
        default="True",
    )

    definition.define_enum_parameter(
        DATA_RETENTION_TIME,
        label="Pod Retention Time in Hours",
        description="The amount of time Pod objects should be retained with a 'Not Existing' or 'None' status before being deleted.",
        values=["02", "06", "12", "24", PLATFORM_DEFAULT],
        default=PLATFORM_DEFAULT,
        required=True,
        advanced=True,
    )

    definition.define_int_parameter(
        QUERY_TIMEOUT,
        label="Query Timeout (s)",
        description="Sets the request timeout in seconds for querying VMware Aria Operations for Applications. Using a timeout greater than 300 seconds will have no effect, as queries that run longer than 300 seconds are stopped by VMware Aria Operations for Applications.",
        required=True,
        advanced=True,
        default=30,
    )

    definition.define_enum_parameter(
        ENABLE_EXTERNAL_LINKS,
        values=["True", "False"],
        label="Enable External Links",
        description="Determines whether this adapter instance should manage external links in VMware Aria Operations for Applications.",
        required=True,
        advanced=True,
        default="True",
    )

    # The key 'container_memory_limit' is a special key that is read by the VMware
    # Aria Operations collector to determine how much memory to allocate to the docker
    # container running this adapter. It does not need to be read inside the adapter
    # code.
    definition.define_int_parameter(
        "container_memory_limit",
        label="Adapter Memory Limit (MB)",
        description="Sets the maximum amount of memory VMware Aria Operations can "
        "allocate to the container running this adapter instance.",
        required=True,
        advanced=True,
        default=1024,
    )

    definition.add_object_type(Cluster.definition())
    definition.add_object_type(Node.definition())
    definition.add_object_type(Namespace.definition())
    definition.add_object_type(Deployment.definition())
    definition.add_object_type(ReplicaSet.definition())
    definition.add_object_type(Pod.definition())

    logger.info("Finished getAdapterDefinition")
    logger.debug(f"Returning definition {definition.to_json()}")

    return definition


# Main entry point of the adapter. You should not need to modify anything below this
# line.
def main(argv):
    logging.setup_logging("adapter.log")
    logging.rotate()
    logger.debug(f"Running adapter code with arguments: {argv}")
    if len(argv) != 3:
        # `inputfile` and `outputfile` are always automatically appended to the
        # argument list by the server
        logger.debug("Arguments must be <method> <inputfile> <ouputfile>")
        exit(1)

    method = argv[0]

    try:
        if method == "test":
            with Timer(logger, "Test"):
                asyncio.run(test(AdapterInstance.from_input())).send_results()
        elif method == "endpoint_urls":
            with Timer(logger, "Get endpoints"):
                get_endpoints(AdapterInstance.from_input()).send_results()
        elif method == "collect":
            with Timer(logger, "Collect"):
                asyncio.run(collect(AdapterInstance.from_input())).send_results()
        elif method == "adapter_definition":
            with Timer(logger, "Adapter definition"):
                result = get_adapter_definition()
                if type(result) is AdapterDefinition:
                    result.send_results()
                else:
                    logger.info(
                        "get_adapter_definition method did not return an AdapterDefinition"
                    )
                    exit(1)
        else:
            logger.debug(f"Command {method} not found")
            exit(1)
    except Exception as e:
        logger.error(f"Adapter exception during '{method}': {e}")
        exit(1)
    logger.info(Timer.graph())


if __name__ == "__main__":
    main(sys.argv[1:])
