import logging
from typing import Dict
from typing import Optional

import cluster
import deployment
import namespace
import node
import pod
import replica_set
from aoa_client import AOAClient
from aria.ops.event import Criticality
from aria.ops.object import Key
from aria.ops.result import CollectResult

logger = logging.getLogger(__name__)


async def add_alerts(
    aoa_client: AOAClient, collect_result: CollectResult, collect_pods
):
    try:
        criticalities = {
            "SEVERE": Criticality.CRITICAL,
            "WARN": Criticality.IMMEDIATE,
            "SMOKE": Criticality.WARNING,
            "INFO": Criticality.INFO,
            "NONE": Criticality.NONE,
        }

        search = {
            #  Note: This response is paged if the number of entries exceeds
            #  'limit'. If we switch to allowing user-created alerts, we will want to
            #  add paging support to post requests.
            "limit": 25,  # We are only expecting 10, use 25 in case more are added
            "offset": 0,
            "query": [
                # To get all alerts with a given tag, query for that tag in the
                #  'tags.customerTags' key. This query searches for any alert
                #  with the tag '~integration.kubernetes', which is the tag for the
                #  default K8s integration alerts.
                # Note: Tags starting with '~' are not allowed to be created or set
                #  by the user. If we want the user to be able to add synced alerts by
                #  tagging an alert, we cannot use '~integration.kubernetes'.
                {
                    "key": "tags.customerTags",
                    "value": "~integration.kubernetes",
                    "matchingMethod": "CONTAINS",
                    "negated": False,
                },
                {
                    "key": "status",
                    "value": "FIRING",
                    "matchingMethod": "EXACT",
                    "negated": False,
                    "start": str(aoa_client.start_time),
                    "end": str(aoa_client.end_time),
                },
            ],
            "sort": {"ascending": True, "field": "updated"},
        }

        result = await aoa_client.post("/api/v2/search/alert", body=search)
        alerts = result.get("response", {}).get("items", [])

        logger.debug(f"Found {len(alerts)} firing alerts")

        for alert in alerts:
            name = alert.get("name")
            logger.debug(f"Processing alert {name}")
            events = alert.get("failingHostLabelPairs", {})
            for event in events:
                try:
                    logger.debug(
                        f"Processing HostLabel pair for firing event: {repr(event)}"
                    )
                    tags = event.get("tags", {})
                    key = _get_key_from_tags(
                        tags, aoa_client.aria_apps_url, collect_pods
                    )

                    if not key:
                        logger.warning(
                            f"Cannot construct a key for event '{repr(event)}'."
                        )
                        continue
                    criticality = criticalities.get(
                        event.get("severity"), Criticality.NONE
                    )
                    start_time = event.get("startTime")

                    obj = collect_result.get_object(key)
                    if obj:
                        obj.with_event(
                            message=name,
                            start_date=start_time,
                            criticality=criticality,
                            # Because VMware Aria Operations for Applications already has a wait cycle built in
                            # for starting/canceling, we should set the VMware Aria Ops
                            # values to 1 so that we don't unnecessarily delay
                            # firing/resolving.
                            watch_wait_cycle=1,
                            cancel_wait_cycle=1,
                        )
                        logger.debug(f"Added event to object {key}: {name}")
                    else:
                        logger.warning(
                            f"Found event on K8s Object with key {key}, but there is "
                            f"no object with that key."
                        )
                except Exception as e:
                    logger.warning(f"Error adding '{name}' alerts to K8s Object: {e}")
    except Exception as e:
        logger.warning(f"Error adding alerts to K8s Object: {e}")


def _get_key_from_tags(tags: Dict, aria_apps_url: str, collect_pods) -> Optional[Key]:
    deployment = tags.get("deployment")
    replicaset = tags.get("replicaset")
    namespace = tags.get("namespace_name")
    node = tags.get("nodename")
    cluster = tags.get("cluster")
    # if not collecting pods, set pod to None so that pod alerts will roll up to a parent
    if collect_pods:
        pod = tags.get("pod_name")
    else:
        pod = None
    return _get_key(
        deployment, replicaset, namespace, pod, node, cluster, aria_apps_url
    )


def _get_key(
    deployment_name: Optional[str],
    replicaset_name: Optional[str],
    namespace_name: Optional[str],
    pod_name: Optional[str],
    node_name: Optional[str],
    cluster_name: Optional[str],
    aria_apps_url: str,
) -> Optional[Key]:
    if cluster_name:
        if namespace_name:
            if pod_name:
                return pod.get_key(
                    cluster_name, namespace_name, pod_name, aria_apps_url
                )
            elif replicaset_name:
                return replica_set.get_key(
                    cluster_name, namespace_name, replicaset_name, aria_apps_url
                )
            elif deployment_name:
                return deployment.get_key(
                    cluster_name, namespace_name, deployment_name, aria_apps_url
                )
            else:
                return namespace.get_key(cluster_name, namespace_name, aria_apps_url)
        if node_name:
            return node.get_key(cluster_name, node_name, aria_apps_url)
        else:
            return cluster.get_key(cluster_name, aria_apps_url)
    else:
        # Could not find an object to send the alert to
        return None
