import logging
import time
from typing import Optional
from urllib import parse

import constants
import httpx
from aria.ops.adapter_instance import AdapterInstance
from constants import ARIA_APPS_CREDENTIAL_KEY
from constants import ARIA_APPS_TOKEN_KEY
from constants import CSP_TOKEN_KEY
from constants import CSP_URL

logger = logging.getLogger(__name__)


def format_url(url: Optional[str]) -> Optional[str]:
    if url is None:
        return url
    # Strip trailing slashes for consistency
    url = url.rstrip("/")
    # Ensure protocol is set
    if "https://" not in url:
        url = f"https://{url}"
    return url


class RestRequestException(Exception):
    def __init__(self, message, status_code):
        super().__init__(message)
        self.status_code = status_code


class AOAClient:
    def __init__(self, adapter_instance: AdapterInstance):
        self.aria_apps_url = (
            format_url(adapter_instance.get_identifier_value(constants.ARIA_APPS_URL))
            or ""
        )

        self.token = _resolve_token(adapter_instance)
        adapter_instance.get_credential_value(constants.ARIA_APPS_TOKEN_KEY)
        if adapter_instance.collection_window:
            self.start_time = round(adapter_instance.collection_window["start_time"])
            self.end_time = round(adapter_instance.collection_window["end_time"])
        else:
            # Fallback when no window is present, e.g., in test connection
            self.end_time = round(time.time() * 1000)
            self.start_time = 0

        # For the first collection, start 5 minutes back from the current collection
        # time.
        if self.start_time == 0:
            self.start_time = self.end_time - (5 * 60000)

        self.timeout = float(
            adapter_instance.get_identifier_value(constants.QUERY_TIMEOUT, "30.0")
        )
        if self.timeout <= 0:
            logger.warning(
                "Negative or zero timeouts are not allowed. Using default 30s timeout."
            )
            self.timeout = 30.0

    async def __aenter__(self):
        """Acquire session upon entering the 'with' context

        :return: self
        """
        self.client = httpx.AsyncClient()
        self.token = await self.token
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Release the token upon exiting the 'with' context

        :param exception_type: Unused
        :param exception_value: Unused
        :param traceback: Unused
        :return: None
        """
        await self.client.aclose()

    async def run_query(self, query):
        request = (
            "/api/v2/chart/api?"
            + f"s={self.start_time}"
            + f"&q={parse.quote_plus(query)}"
            + f"&e={self.end_time}"
            + "&g=m"
        )
        return await self.get(request)

    async def get(self, uri):
        return await self._request(uri, self.client.get)

    async def get_paged(self, uri, offset, limit):
        request = uri + f"?offset={offset}" + f"&limit={limit}"
        result = await self.get(request)
        all_items = result.get("response", {}).get("items", [])
        while result.get("response", {}).get("moreItems", False):
            offset += limit
            request = uri + f"?offset={offset}" + f"&limit={limit}"
            result = await self.get(request)
            all_items.extend(result.get("response", {}).get("items", []))
        return all_items

    async def post(self, uri, body):
        return await self._request(uri, self.client.post, json=body)

    async def delete(self, uri):
        return await self._request(uri, self.client.delete)

    async def _request(self, uri, request_fn, **kwargs):
        try:
            url = self.aria_apps_url

            # VMware Aria Operations for Applications does not handle double slashes in
            # its urls. We could do a global find/replace for "//" -> "/" but there may
            # be a valid reason for those to exist in part of the url, e.g., a query.
            uri = uri.lstrip("/")
            request = f"{url}/{uri}"

            response = await request_fn(
                request, headers=self.create_headers(), timeout=self.timeout, **kwargs
            )
            if response.status_code < 300:
                logger.debug(
                    f"request <{repr(request)}> responded with {repr(response.status_code)}"
                )
                return response.json()
            else:
                logger.warning(
                    f"request <{repr(request)}> responded with {repr(response.status_code)}"
                )
                raise RestRequestException(response.text, response.status_code)
        except RestRequestException as e:
            raise e
        except Exception as e:
            logger.warning(
                f"failed to execute {request_fn.__name__.upper()} request {uri}"
            )
            logger.exception(e)
            raise e

    def create_headers(self):
        return {"accept": "application/json", "Authorization": "Bearer " + self.token}


async def _resolve_token(adapter_instance: AdapterInstance) -> str:
    logger.debug("Resolving Token")
    token = ""

    if adapter_instance.get_credential_type() == ARIA_APPS_CREDENTIAL_KEY:
        token = adapter_instance.get_credential_value(ARIA_APPS_TOKEN_KEY)
    else:
        token = await _authorize_api_tokens(
            adapter_instance.get_credential_value(CSP_TOKEN_KEY)
        )

    return token


async def _authorize_api_tokens(csp_token) -> str:
    logger.debug("Obtaining CSP Token")

    headers = {
        "accept": "application/json",
        "Content-Type": "application/x-www-form-urlencoded",
    }
    data = {
        "api_token": csp_token,
    }

    async with httpx.AsyncClient() as client:
        response = await client.post(CSP_URL, headers=headers, data=data)

    if response.status_code == 200:
        return response.json().get("access_token")
    else:
        raise Exception(
            f"Request failed with status code {response.status_code}: {response.text}"
        )
