import asyncio
from types import FunctionType
from typing import AsyncIterator
from typing import Generator
from typing import Sequence
from typing import Union

from aoa_client import AOAClient
from aria.ops.object import Key
from aria.ops.object import Metric
from aria.ops.object import Object
from aria.ops.object import Property
from aria.ops.pipe_utils import logging

logger = logging.getLogger(__name__)


async def get_timeseries(
    aoa_client: AOAClient, queries: Sequence[tuple]
) -> AsyncIterator[tuple]:
    tasks = []
    keys = []
    functions = []
    for key, query, fn in queries:
        logger.debug(f"Running request for {key}")
        tasks.append(asyncio.ensure_future(aoa_client.run_query(query)))
        keys.append(key)
        functions.append(fn)

    responses = await asyncio.gather(*tasks, return_exceptions=True)

    for key, response, fn in zip(keys, responses, functions):
        try:
            # Since there can be Exceptions during the run_query execution, the response might be an exception
            if isinstance(response, Exception):
                raise response
            elif "errorMessage" in response:
                logger.error(
                    f"Error while fetching '{key}' metric: {response['errorMessage']}"
                )
            elif "warnings" in response:
                logger.debug(f"Warning while fetching '{key}': {response['warnings']}")

            if "timeseries" not in response:
                logger.warning(f"No timeseries found in response for '{key}'")
                logger.debug(f"skipping '{key}'")
                continue

            yield key, response["timeseries"], fn
        except Exception as e:
            logger.error(f"Unable to produce timeseries for '{key}' due to: {e}")


def add_attribute(object: Object, attribute: Union[Metric, Property]):
    logger.debug(f"Adding attribute '{attribute.key}' to {object.get_key().name}")
    if type(attribute) is Metric:
        object.add_metric(attribute)
    else:
        object.add_property(attribute)


def generate_metric(
    metric_key: str,
    timeseries: dict,
    aria_apps_url: str,
    get_key_from_tags: FunctionType,
) -> Generator[tuple[Key, Metric], None, None]:
    logger.debug(f"Processing timeseries data for metric '{metric_key}': {timeseries}")
    for ts in timeseries:
        tags = ts.get("tags", {})
        key = get_key_from_tags(tags, aria_apps_url)
        if ts["data"]:
            last_value_in_timeseries = ts["data"][-1][1]
            if key:
                yield key, Metric(metric_key, last_value_in_timeseries)
        else:
            logger.warning(f"No data for '{metric_key}' on {key}")


def generate_property(
    property_key: str,
    timeseries: dict,
    aria_apps_url: str,
    get_key_from_tags: FunctionType,
) -> Generator[tuple[Key, Property], None, None]:
    logger.debug(
        f"Processing timeseries data for property '{property_key}': {timeseries}"
    )
    for ts in timeseries:
        tags = ts.get("tags", {})
        key = get_key_from_tags(tags, aria_apps_url)
        attribute_value = tags[property_key]
        if key:
            yield key, Property(property_key, attribute_value)


def generate_properties(
    property_keys: list[str],
    timeseries: dict,
    aria_apps_url: str,
    get_key_from_tags: FunctionType,
) -> Generator[tuple[Key, Property], None, None]:
    logger.debug(
        f"Processing timeseries data for attribute '{property_keys}': {timeseries}"
    )
    for ts in timeseries:
        tags = ts.get("tags", {})
        key = get_key_from_tags(tags, aria_apps_url)

        for property_key in property_keys:
            logger.debug(f"property key: {property_key}")
            attribute_value = tags.get(property_key, None)

            if attribute_value and key:
                yield key, Property(property_key, attribute_value)
            else:
                logging.debug(
                    f"No property '{property_key}' value found for '{key.name}'"
                )
