import logging
from typing import Optional

import constants
from aoa_client import AOAClient
from aria.ops.definition.object_type import ObjectType
from aria.ops.definition.units import Units
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from attribute_utils import add_attribute
from attribute_utils import generate_metric
from attribute_utils import get_timeseries
from error_handler import safe_call

logger = logging.getLogger(__name__)


class Cluster(Object):
    def __init__(self, cluster, aria_apps_url):
        super().__init__(get_key(cluster, aria_apps_url))

    @staticmethod
    def definition():
        cluster_definition = ObjectType(constants.CLUSTER_OBJECT_KIND, "K8s Cluster")

        # Identifiers
        cluster_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster Name"
        )
        cluster_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        cluster_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Metrics
        cluster_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_ACTIVE_NODES_KEY, "Active Nodes"
        )
        cluster_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_INACTIVE_NODES_KEY, "Inactive Nodes"
        )
        cluster_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_NAMESPACES_KEY, "Namespaces"
        )
        cluster_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY, "Pods"
        )
        cluster_definition.define_metric(
            constants.METRIC_CPU_USAGE_KEY, "CPU Usage"
        )  # Units millicores
        cluster_definition.define_metric(
            constants.METRIC_MEMORY_USAGE_KEY, "Memory Usage", Units.DATA_SIZE.BIBYTE
        )
        cluster_definition.define_metric(
            constants.METRIC_AVERAGE_MEMORY_UTILIZATION_KEY,
            "Average Memory Utilization",
            Units.RATIO.PERCENT,
        )
        cluster_definition.define_metric(
            constants.METRIC_AVERAGE_CPU_UTILIZATION_KEY,
            "Average CPU Utilization",
            Units.RATIO.PERCENT,
        )
        cluster_definition.define_metric(
            constants.METRIC_AVERAGE_STORAGE_UTILIZATION_KEY,
            "Average Storage Utilization",
            Units.RATIO.PERCENT,
        )

        return cluster_definition


@safe_call
async def get_clusters(
    aoa_client: AOAClient, metric_prefixes: list[str]
) -> dict[Key, Cluster]:
    queries_and_metrics_keys = []
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        queries_and_metrics_keys.extend(
            [
                (
                    constants.METRIC_TOTAL_NUMBER_OF_ACTIVE_NODES_KEY,
                    f"at(5m,align(5m,count(ts({metric_prefix}.node.memory.working_set), cluster)))",
                    generate_metric,
                ),
                (
                    constants.METRIC_TOTAL_NUMBER_OF_INACTIVE_NODES_KEY,
                    f"at(5m,align(5m, count(lowpass(1, ts({metric_prefix}.node.status.condition,condition='Ready')),cluster)))",
                    generate_metric,
                ),
                (
                    # TODO: This one is a bit slow
                    constants.METRIC_TOTAL_NUMBER_OF_NAMESPACES_KEY,
                    f"at(5m,count(align(5m, mean, count(ts({metric_prefix}.ns.memory.usage), cluster, namespace_name)), cluster))",
                    generate_metric,
                ),
                (
                    constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY,
                    f"at(5m,sum(align(5m, mean, ts({metric_prefix}.cluster.pod.count)),cluster))",
                    generate_metric,
                ),
                (
                    constants.METRIC_CPU_USAGE_KEY,
                    f"at(5m,align(5m,sum(ts({metric_prefix}.cluster.cpu.usage_rate), cluster)))",
                    generate_metric,
                ),
                (
                    constants.METRIC_MEMORY_USAGE_KEY,
                    f"at(5m,align(5m,sum(ts({metric_prefix}.cluster.memory.usage),cluster)))",
                    generate_metric,
                ),
                (
                    constants.METRIC_AVERAGE_MEMORY_UTILIZATION_KEY,
                    f"at(5m,avg(align(5m,ts({metric_prefix}.node.memory.working_set)) / align(5m,ts({metric_prefix}.node.memory.node_allocatable)) * 100, cluster))",
                    generate_metric,
                ),
                (
                    constants.METRIC_AVERAGE_CPU_UTILIZATION_KEY,
                    f"at(5m,avg(align(5m,mean,ts({metric_prefix}.node.cpu.node_utilization)), cluster) * 100)",
                    generate_metric,
                ),
                (
                    constants.METRIC_AVERAGE_STORAGE_UTILIZATION_KEY,
                    f"at(5m,avg(align(5m,ts({metric_prefix}.node.filesystem.usage)) / align(5m,ts({metric_prefix}.node.filesystem.limit)),cluster) * 100)",
                    generate_metric,
                ),
            ]
        )

    clusters: dict[Key, Cluster] = {}
    async for attribute_key, timeseries, attribute_generating_function in get_timeseries(
        aoa_client, queries_and_metrics_keys
    ):
        try:
            for object_key, attribute in attribute_generating_function(
                attribute_key, timeseries, aoa_client.aria_apps_url, _get_key_from_tags
            ):
                # Add metric value to dict of metric values
                if object_key not in clusters:
                    new_cluster = Cluster(
                        object_key.get_identifier(constants.IDENTIFIER_CLUSTER_KEY),
                        aoa_client.aria_apps_url,
                    )
                    new_cluster.with_property(
                        constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                    )
                    clusters[object_key] = new_cluster
                    logger.debug(f"Created new cluster '{object_key}'.")

                add_attribute(clusters[object_key], attribute)
        except Exception as e:
            logger.warning(
                f"Unable to add metric with key '{attribute_key}' to cluster due to: {e}"
            )
            logger.exception(e)

    for cluster in clusters.values():
        active_nodes = cluster.get_last_metric_value(
            constants.METRIC_TOTAL_NUMBER_OF_ACTIVE_NODES_KEY
        )
        inactive_nodes = cluster.get_last_metric_value(
            constants.METRIC_TOTAL_NUMBER_OF_INACTIVE_NODES_KEY
        )

        if inactive_nodes is None and active_nodes is not None:
            cluster.with_metric(constants.METRIC_TOTAL_NUMBER_OF_INACTIVE_NODES_KEY, 0)
        else:
            logger.debug(
                f"Skipping default metric value '{constants.METRIC_TOTAL_NUMBER_OF_INACTIVE_NODES_KEY}' for cluster {cluster}. {constants.METRIC_TOTAL_NUMBER_OF_ACTIVE_NODES_KEY}: {active_nodes}"
            )

    logger.info(f"Collected {len(clusters)} total clusters")
    return clusters


# Helpers
def _get_key_from_tags(tags, aria_apps_url) -> Optional[Key]:
    cluster = tags.get("cluster")
    return get_key(cluster, aria_apps_url)


def get_key(cluster: Optional[str], aria_apps_url: str) -> Optional[Key]:
    if not cluster:
        return None

    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.CLUSTER_OBJECT_KIND,
        name=cluster,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )
