import logging
from typing import Optional

import constants
import namespace
from aoa_client import AOAClient
from aria.ops.definition.object_type import ObjectType
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from error_handler import safe_call
from namespace import Namespace
from relationships import add_relationships

logger = logging.getLogger(__name__)


class Deployment(Object):
    def __init__(self, cluster, namespace, deployment, aria_apps_url):
        super().__init__(get_key(cluster, namespace, deployment, aria_apps_url))

    @staticmethod
    def definition():
        deployment_definition = ObjectType(
            constants.DEPLOYMENT_OBJECT_KIND, "K8s Deployment"
        )

        # Identifiers
        deployment_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster"
        )
        deployment_definition.define_string_identifier(
            constants.IDENTIFIER_NAMESPACE_KEY, "Namespace"
        )
        deployment_definition.define_string_identifier(
            constants.IDENTIFIER_DEPLOYMENT_KEY, "Deployment Name"
        )
        deployment_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        deployment_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        return deployment_definition


@safe_call
async def get_deployments(
    aoa_client: AOAClient, metric_prefixes
) -> dict[Key, Deployment]:
    deployments = {}
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        logger.info(f"Collecting deployments for metric prefix '{metric_prefix}'")
        try:
            response = await aoa_client.run_query(
                f"at(5m,rawsum(ts({metric_prefix}.deployment.available_replicas), cluster, namespace_name, deployment))",
            )
        except Exception as e:
            logger.exception(e)
            continue
        if "errorMessage" in response:
            logger.error(
                f"Error while fetching deployments: {response['errorMessage']}"
            )
            continue
        elif "warnings" in response:
            logger.warning(
                f"Warning while fetching deployments: {response['warnings']}"
            )

        if "timeseries" not in response:
            logger.warning(
                f"No deployments returned for metric prefix '{metric_prefix}'"
            )
            continue

        new_count = 0
        for deployment_timeseries in response["timeseries"]:
            try:
                tags = deployment_timeseries["tags"]

                cluster = tags.pop("cluster")
                namespace = tags.pop("namespace_name")
                deployment = tags.pop("deployment")

                new_deployment = Deployment(
                    cluster, namespace, deployment, aoa_client.aria_apps_url
                )
                key = new_deployment.get_key()

                if key in deployments:
                    logger.warning(
                        f"Duplicate Deployment key '{key}' was found during collection"
                    )
                    continue

                new_deployment.with_property(
                    constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                )
                deployments[key] = new_deployment

                logger.debug(f"Created new deployment '{key}'.")
                new_count += 1

            except Exception as e:
                logger.warning(f"Unable to add data to a Deployment due to: {e}")
                logger.exception(e)
        logger.info(
            f"Collected {new_count} deployments for metric prefix '{metric_prefix}'"
        )
    logger.info(f"Collected {len(deployments)} total deployments")
    return deployments


def add_deployment_relationships(
    deployments: dict[Key, Deployment], namespaces: dict[Key, Namespace]
) -> None:
    def get_namespace_key(deployment: Deployment) -> Optional[Key]:
        return namespace.get_key(
            deployment.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            deployment.get_identifier_value(constants.IDENTIFIER_NAMESPACE_KEY),
            deployment.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    add_relationships(namespaces, deployments, logger, get_namespace_key)


def get_key(
    cluster: Optional[str],
    namespace: Optional[str],
    deployment: Optional[str],
    aria_apps_url: str,
) -> Optional[Key]:
    if not cluster or not namespace or not deployment:
        return None

    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.DEPLOYMENT_OBJECT_KIND,
        name=deployment,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_NAMESPACE_KEY, value=namespace),
            Identifier(key=constants.IDENTIFIER_DEPLOYMENT_KEY, value=deployment),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )
