import json
import logging

import constants
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from aria.ops.suite_api_client import SuiteApiClient
from node import Node

logger = logging.getLogger(__name__)


class EC2Instance(Object):
    def __init__(self, ec2_instance_name, identifiers) -> None:
        super().__init__(
            Key(
                adapter_kind=constants.AWS_ADAPTER_KIND,
                object_kind=constants.AWS_EC2_KIND,
                name=ec2_instance_name,
                identifiers=identifiers,
            )
        )


def get_ec2_instances(
    client: SuiteApiClient, nodes: dict[Key, Node]
) -> list[EC2Instance]:
    ec2_instances = []
    # for each need, build a query to fetch the ec2 instances with the given node's
    # provider_id
    for key, node in nodes.items():
        try:
            # This should always exist because we at least have the default value, 'Unknown'
            provider_id = node.get_last_property_value(
                constants.PROPERTY_PROVIDER_ID_KEY
            )

            # skip any non-aws nodes
            if "aws" not in provider_id:
                logger.debug(f"Skipping non-AWS node with provider_id: {provider_id}")
                continue

            logger.debug(
                f"Trying to find EC2 Instance corresponding to AWS node with provider "
                f"ID: {provider_id}"
            )

            # provider_id looks like this: aws:///us-east-1a/i-06c16e82572aac6b1
            # we only want to look for the last part, which corresponds to the AWS MP
            # EC2 Instance object name: i-06c16e82572aac6b1
            provider_id = provider_id.rsplit("/", 1)[-1]

            logger.debug(f"Searching for EC2 Instance with provider ID: {provider_id}")
            url = f"/api/adapterkinds/{constants.AWS_ADAPTER_KIND}/resourcekinds/{constants.AWS_EC2_KIND}/resources?identifiers[InstanceId]={provider_id}"

            response = client.get(url=url)
            logger.debug(
                f"request {repr(url)} responded w/" + repr(response.status_code)
            )

            if response.status_code >= 300:
                logger.warning(
                    "skipping collection for ec2 instance due to unsuccessful server "
                    "response"
                )
                continue

            content = response.content.decode("utf-8")
            parsed = json.loads(content)

            if "resourceList" not in parsed or not parsed["resourceList"]:
                logger.debug(
                    f"Failed to find an EC2 instance object for AWS node with provider "
                    f"ID: {provider_id}"
                )
                continue

            ec2_instance = parsed["resourceList"][0]
            # create the resource
            name = ec2_instance["resourceKey"]["name"]

            identifiers = []
            for identifier in ec2_instance["resourceKey"]["resourceIdentifiers"]:
                identifiers.append(
                    Identifier(
                        identifier["identifierType"]["name"],
                        identifier["value"],
                        identifier["identifierType"]["isPartOfUniqueness"],
                    )
                )

            new_ec2_instance = EC2Instance(name, identifiers)
            new_ec2_instance.add_child(node)

            ec2_instances.append(new_ec2_instance)

        except Exception as e:
            logger.error(
                f"Suite API collection failed while trying to fetch EC2 Instance "
                f"related to {key.name}"
            )
            logger.exception(e)

    return ec2_instances
