import json
import logging

import constants

EXT_LINK_PREFIX = "Troubleshooting Workbench - "

logger = logging.getLogger(__name__)


async def sync_external_links(
    aoa_client, suite_api_client, aria_apps_url, aria_ops_url
):
    """
    Manage external links in AOA. Automatically create missing links and delete
    stale links based on K8s Node objects received from the VMware Aria Operations
    Suite API.
    """
    # get current "data receiving" nodes from AO
    try:
        ao_nodes = get_ao_nodes(suite_api_client, aria_apps_url)
    except Exception as e:
        logger.error(
            "Unable to retrieve Node objects from VMware Aria Operations Suite API"
        )
        logger.exception(e)
        return

    # get current external links from AOA
    try:
        external_links = await get_external_links(aoa_client)
    except Exception as e:
        logger.error(
            "Unable to retrieve external links from VMware Aria Operations for "
            "Applications API"
        )
        logger.exception(e)
        return

    # check for missing external links
    for node_uuid in ao_nodes:
        cluster_name = ao_nodes[node_uuid][0]
        node_name = ao_nodes[node_uuid][1]
        # create external links that don't already exist
        logger.debug(
            f"Checking for external link for {node_name} in cluster {cluster_name} with UUID {node_uuid}"
        )
        if node_uuid not in external_links:
            logger.info(
                f"External link not found for {node_name}. Creating external link."
            )
            await create_external_link(
                aoa_client, aria_ops_url, node_name, cluster_name, node_uuid
            )
        else:
            logger.debug("External link found.")

    # check for stale external links
    for external_link_uuid in external_links:
        ext_link_id = external_links[external_link_uuid][0]
        ext_link_name = external_links[external_link_uuid][1]
        if external_link_uuid not in ao_nodes:
            # delete stale external link
            await delete_external_link(aoa_client, ext_link_name, ext_link_id)


async def create_external_link(
    aoa_client, aria_ops_url, node_name, cluster_name, node_uuid
):
    logger.info(
        f"Creating external link for node {node_name} in cluster {cluster_name} with UUID {node_uuid}"
    )
    ext_link_body = {
        "name": f"{EXT_LINK_PREFIX}{node_name}",
        "description": f"Auto-generated external link for Kubernetes Node {node_name} from Cluster {cluster_name}",
        "pointTagFilterRegexes": {
            "nodename": node_name,
            "cluster": cluster_name,
        },
        "isLogIntegration": "false",
        "template": f"{aria_ops_url}/ui/index.action#troubleshoot/workbench/{node_uuid}/root-cause",
    }
    try:
        await aoa_client.post("/api/v2/extlink", ext_link_body)
    except Exception as e:
        logger.error("External link creation failed")
        logger.exception(e)
    logger.info("External link creation succeeded")


async def get_external_links(aoa_client):
    # get the current external links from AOA
    # NOTE: errors must be handled by caller
    current_ext_links = await aoa_client.get_paged(
        "/api/v2/extlink", offset=0, limit=100
    )
    logger.info(f"Retrieved {len(current_ext_links)} external links")

    # build a map of external link UUIDs to IDs, names, and URLs
    external_links = {}
    for current_ext_link in current_ext_links:
        try:
            ext_link_id = current_ext_link.get("id")
            ext_link_name = current_ext_link.get("name")
            ext_link_url = current_ext_link.get("template")
            ext_link_uuid = ext_link_url.split("/")[-2]
        except Exception as e:
            # log the error but keep trying to sync other external links
            logger.warn("Unable to parse name, ID, and/or URL for external node")
            logger.exception(e)
            continue

        # If it doesn't start with our prefix, we didn't make it; just skip it.
        if EXT_LINK_PREFIX not in ext_link_name:
            logger.debug(f'Ignoring unrelated external link: "{ext_link_name}"')
            continue

        external_links[ext_link_uuid] = (ext_link_id, ext_link_name, ext_link_url)

    return external_links


def get_ao_nodes(suite_api_client, aria_apps_url):
    # we need to get nodes from VMware Aria Operations (via Suite API) because we need
    # the VMware Aria Operations object UUIDs which are part of the External Link URLs
    nodes = get_nodes_from_suite_api(suite_api_client, aria_apps_url)

    # build a map of nodes to boolean to determine if we've seen it already
    # build a map of UUIDs to names for external links
    ao_nodes = {}
    for node in nodes:
        try:
            uuid = node["identifier"]
            node_name = node["resourceKey"]["name"]
            cluster_name = parse_cluster(node)
            ao_nodes[uuid] = (cluster_name, node_name)
        except Exception as e:
            # log the error but keep trying to parse other nodes
            logger.warn("Unable to parse UUID, name, or cluster for node")
            logger.exception(e)
    return ao_nodes


def get_nodes_from_suite_api(suite_api_client, aria_apps_url):
    """
    Fetches K8s Node objects from Suite API.  Only gets nodes that are receiving
    data and are collected by this adapter instance (based on the aria_apps_url)
    """
    query = {
        "adapterKind": [constants.ADAPTER_KIND],
        "resourceKind": [constants.NODE_OBJECT_KIND],
        "resourceStatus": ["DATA_RECEIVING"],
        "propertyConditions": {
            "conditions": [
                {
                    "key": "aria_apps_url",
                    "operator": "EQ",
                    "stringValue": aria_apps_url,
                }
            ]
        },
    }
    response = suite_api_client.post(url="/api/resources/query", json=query)
    logger.info(f"Request {repr(query)} responded: " + repr(response.status_code))
    if response.status_code >= 300:
        logger.info("Failed to retrieve nodes from Suite API")
        return []
    content = response.content.decode("utf-8")
    parsed = json.loads(content)
    if not len(parsed.get("resourceList", [])):
        logger.warning("No Nodes found from Suite API")
        return []
    return parsed["resourceList"]


async def delete_external_link(aoa_client, ext_link_name, ext_link_id):
    logger.info(
        f'Deleting stale external link: "{ext_link_name}" with id {ext_link_id}'
    )
    try:
        await aoa_client.delete(f"/api/v2/extlink/{ext_link_id}")
    except Exception as e:
        # if deletion fails just log and move on
        logger.error(f'Failed to delete stale external link: "{ext_link_name}".')
        logger.exception(e)


def parse_cluster(node):
    if "resourceKey" not in node or "resourceIdentifiers" not in node["resourceKey"]:
        return ""

    for identifier in node["resourceKey"]["resourceIdentifiers"]:
        if (
            "identifierType" not in identifier
            or "value" not in identifier
            or "name" not in identifier["identifierType"]
        ):
            continue

        if identifier["identifierType"]["name"] == "cluster":
            return identifier["value"]

    return ""
