import logging
from typing import Optional

import cluster
import constants
from aoa_client import AOAClient
from aria.ops.definition.object_type import ObjectType
from aria.ops.definition.units import Units
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from attribute_utils import add_attribute
from attribute_utils import generate_metric
from attribute_utils import get_timeseries
from cluster import Cluster
from error_handler import safe_call
from relationships import add_relationships

logger = logging.getLogger(__name__)


class Namespace(Object):
    def __init__(self, cluster, namespace, aria_apps_url):
        super().__init__(get_key(cluster, namespace, aria_apps_url))

    @staticmethod
    def definition() -> ObjectType:
        namespace_definition = ObjectType(
            constants.NAMESPACE_OBJECT_KIND, "K8s Namespace"
        )

        # Identifiers
        namespace_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster"
        )
        namespace_definition.define_string_identifier(
            constants.IDENTIFIER_NAMESPACE_KEY, "Namespace"
        )
        namespace_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        namespace_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )
        # Metrics
        namespace_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY, "Total Number of Pods"
        )
        namespace_definition.define_metric(
            constants.METRIC_POD_STATE_KEY_RUNNING, "Number of pods with status Running"
        )
        namespace_definition.define_metric(
            constants.METRIC_POD_STATE_KEY_PENDING, "Number of pods with status Pending"
        )
        namespace_definition.define_metric(
            constants.METRIC_POD_STATE_KEY_SUCCEEDED,
            "Number of pods with status Succeeded",
        )
        namespace_definition.define_metric(
            constants.METRIC_POD_STATE_KEY_FAILED, "Number of pods with status Failed"
        )
        namespace_definition.define_metric(
            constants.METRIC_POD_STATE_KEY_UNKNOWN, "Number of pods with status Unknown"
        )
        namespace_definition.define_metric(
            constants.METRIC_CPU_UTILIZATION_KEY, "CPU Utilization", Units.RATIO.PERCENT
        )
        namespace_definition.define_metric(
            constants.METRIC_MEMORY_USAGE_KEY,
            "Memory Usage",
            Units.DATA_SIZE.BIBYTE,
        )
        namespace_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_DEPLOYMENTS, "Deployments"
        )

        return namespace_definition


@safe_call
async def get_namespaces(
    aoa_client: AOAClient, metric_prefixes
) -> dict[Key, Namespace]:
    # This query should return any namespace cpu usage rate metrics that have come
    # back in the last 5 minutes. Since the Kubernetes integration for VMware Aria
    # for Applications collects data every few seconds, a 5-minute window should be
    # more than sufficient to capture all the namespaces present in the monitored
    # Kubernetes environment. However, the downside of this approach is that this
    # query returns way more data than we actually need, since we only care about the
    # metadata.
    queries_and_metrics_keys = []
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        queries_and_metrics_keys.extend(
            [
                (
                    constants.METRIC_CPU_UTILIZATION_KEY,
                    f"at(5m,avg(align(5m, ts({metric_prefix}.ns.cpu.usage_rate)), cluster,  namespace_name))",
                    generate_metric,
                ),
                (
                    constants.METRIC_MEMORY_USAGE_KEY,
                    f"at(5m,sum(align(5m,ts({metric_prefix}.ns.memory.usage)),cluster, namespace_name))",
                    generate_metric,
                ),
                (
                    constants.METRIC_TOTAL_NUMBER_OF_DEPLOYMENTS,
                    f"at(5m,count(ts({metric_prefix}.deployment.available_replicas), namespace_name, cluster))",
                    generate_metric,
                ),
            ]
        )

    namespaces: dict[Key, Namespace] = {}
    async for attribute_key, timeseries, attribute_generating_function in get_timeseries(
        aoa_client, queries_and_metrics_keys
    ):
        try:
            for object_key, attribute in attribute_generating_function(
                attribute_key, timeseries, aoa_client.aria_apps_url, _get_key_from_tags
            ):
                # Add metric value to dict of metric values
                if object_key not in namespaces:
                    new_namespace = Namespace(
                        object_key.get_identifier(constants.IDENTIFIER_CLUSTER_KEY),
                        object_key.get_identifier(constants.IDENTIFIER_NAMESPACE_KEY),
                        aoa_client.aria_apps_url,
                    )
                    new_namespace.with_property(
                        constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                    )
                    namespaces[object_key] = new_namespace
                    logger.debug(f"Created new namespace '{object_key}'.")

                add_attribute(namespaces[object_key], attribute)
        except Exception as e:
            logger.warning(
                f"Unable to add metric with key '{attribute_key}' to nodes due to: {e}"
            )
            logger.exception(e)

    logger.info(f"Collected {len(namespaces)} total namespaces")
    return namespaces


# Helpers
def _get_key_from_tags(tags, aria_apps_url) -> Optional[Key]:
    namespace = tags.get("namespace_name")
    cluster = tags.get("cluster")
    return get_key(cluster, namespace, aria_apps_url)


def get_key(
    cluster: Optional[str], namespace: Optional[str], aria_apps_url: str
) -> Optional[Key]:
    if not namespace or not cluster:
        return None

    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.NAMESPACE_OBJECT_KIND,
        name=namespace,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_NAMESPACE_KEY, value=namespace),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )


def add_namespace_relationships(
    namespaces: dict[Key, Namespace], clusters: dict[Key, Cluster]
) -> None:
    def get_cluster_key(namespace: Namespace) -> Optional[Key]:
        return cluster.get_key(
            namespace.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            namespace.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    add_relationships(clusters, namespaces, logger, get_cluster_key)
