import logging
from collections.abc import Generator
from types import FunctionType
from typing import Optional

import cluster
import constants
from aoa_client import AOAClient
from aria.ops.definition.group import Group
from aria.ops.definition.object_type import ObjectType
from aria.ops.definition.units import Units
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from aria.ops.object import Property
from attribute_utils import add_attribute
from attribute_utils import generate_metric
from attribute_utils import generate_properties
from attribute_utils import get_timeseries
from cluster import Cluster
from error_handler import safe_call
from relationships import add_relationships

logger = logging.getLogger(__name__)


class Node(Object):
    def __init__(self, cluster, node, aria_apps_url):
        super().__init__(get_key(cluster, node, aria_apps_url))

    @staticmethod
    def definition():
        node_definition = ObjectType(constants.NODE_OBJECT_KIND, "K8s Node")

        # Identifiers
        node_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster"
        )
        node_definition.define_string_identifier(constants.IDENTIFIER_NODE_KEY, "Node")
        node_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        node_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )
        node_definition.define_string_property(
            constants.PROPERTY_INTERNAL_IP_KEY, "IP Address"
        )
        node_definition.define_string_property(
            constants.PROPERTY_PROVIDER_ID_KEY, "Provider ID"
        )
        node_definition.define_string_property(constants.PROPERTY_NODE_ROLE_KEY, "Role")
        node_definition.define_string_property(
            constants.PROPERTY_NODE_STATUS_KEY, "Status"
        )

        # Metrics
        cpu_group = Group(constants.CPU_GROUP_KEY, "CPU")
        node_definition.add_group(cpu_group)
        cpu_group.define_metric(
            constants.METRIC_CPU_ALLOCATABLE_KEY, "CPU Allocatable"
        )  # NOTE: CPU units are in millicores
        cpu_group.define_metric(constants.METRIC_CPU_CAPACITY_KEY, "CPU Capacity")
        cpu_group.define_metric(constants.METRIC_CPU_RESERVED_KEY, "CPU Reserved")
        cpu_group.define_metric(
            constants.METRIC_CPU_UTILIZATION_KEY, "CPU Utilization", Units.RATIO.PERCENT
        )
        cpu_group.define_metric(
            constants.METRIC_CPU_LIMIT_KEY, "CPU Limit"
        )  # unit = millicores

        memory_group = Group(constants.MEMORY_GROUP_KEY, "Memory")
        node_definition.add_group(memory_group)
        memory_group.define_metric(
            constants.METRIC_MEMORY_ALLOCATABLE_KEY,
            "Memory Allocatable",
            Units.DATA_SIZE.BIBYTE,
        )
        memory_group.define_metric(
            constants.METRIC_MEMORY_CAPACITY_KEY,
            "Memory Capacity",
            Units.DATA_SIZE.BIBYTE,
        )
        memory_group.define_metric(
            constants.METRIC_MEMORY_RESERVED_KEY,
            "Memory Reserved",
            Units.DATA_SIZE.BIBYTE,
        )
        memory_group.define_metric(
            constants.METRIC_MEMORY_UTILIZATION_KEY,
            "Memory Utilization",
            Units.RATIO.PERCENT,
        )

        memory_group.define_metric(
            constants.METRIC_MEMORY_LIMIT_KEY,
            "Memory Limit",
            Units.RATIO.PERCENT,
        )

        filesystem_group = Group(constants.FILESYSTEM_GROUP_KEY, "Filesystem")
        node_definition.add_group(filesystem_group)
        filesystem_group.define_metric(
            constants.METRIC_FILESYSTEM_AVAILABLE_KEY,
            "Filesystem Available",
            Units.DATA_SIZE.BIBYTE,
        )
        filesystem_group.define_metric(
            constants.METRIC_FILESYSTEM_LIMIT_KEY,
            "Filesystem Limit",
            Units.DATA_SIZE.BIBYTE,
        )
        filesystem_group.define_metric(
            constants.METRIC_FILESYSTEM_USAGE_KEY,
            "Filesystem Usage",
            Units.RATIO.PERCENT,
        )
        node_definition.define_metric(
            constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY, "Number of Pods"
        )

        return node_definition


@safe_call
async def get_nodes(aoa_client: AOAClient, metric_prefixes) -> dict[Key, Node]:
    queries_and_metrics_keys = []
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        queries_and_metrics_keys.extend(
            [
                (
                    ["provider_id", "node_role", "internal_ip"],
                    f"at(5m, align(5m,ts({metric_prefix}.node.info)))",
                    generate_properties,
                ),
                (
                    f"{constants.CPU_GROUP_KEY}|{constants.METRIC_CPU_ALLOCATABLE_KEY}",
                    f"at(5m,align(5m, ts({metric_prefix}.node.cpu.node_allocatable)))",
                    generate_metric,
                ),
                (
                    f"{constants.CPU_GROUP_KEY}|{constants.METRIC_CPU_CAPACITY_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.cpu.node_capacity)))",
                    generate_metric,
                ),
                (
                    f"{constants.CPU_GROUP_KEY}|{constants.METRIC_CPU_RESERVED_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.cpu.node_reservation)))",
                    generate_metric,
                ),
                (
                    f"{constants.CPU_GROUP_KEY}|{constants.METRIC_CPU_UTILIZATION_KEY}",
                    f"at(5m,avg(align(5m,ts({metric_prefix}.node.cpu.node_utilization)), cluster, nodename) * 100)",
                    generate_metric,
                ),
                (
                    f"{constants.CPU_GROUP_KEY}|{constants.METRIC_CPU_LIMIT_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.cpu.limit)))",
                    generate_metric,
                ),
                (
                    f"{constants.MEMORY_GROUP_KEY}|{constants.METRIC_MEMORY_ALLOCATABLE_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.memory.node_allocatable)))",
                    generate_metric,
                ),
                (
                    f"{constants.MEMORY_GROUP_KEY}|{constants.METRIC_MEMORY_CAPACITY_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.memory.node_capacity)))",
                    generate_metric,
                ),
                (
                    f"{constants.MEMORY_GROUP_KEY}|{constants.METRIC_MEMORY_RESERVED_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.memory.node_reservation)))",
                    generate_metric,
                ),
                (
                    f"{constants.MEMORY_GROUP_KEY}|{constants.METRIC_MEMORY_UTILIZATION_KEY}",
                    f"""
                    at(5m,avg(
                        align(5m,ts({metric_prefix}.node.memory.working_set))
                        / align(5m,ts({metric_prefix}.node.memory.node_allocatable)) * 100
                    , cluster, nodename))
                    """,
                    generate_metric,
                ),
                (
                    f"{constants.MEMORY_GROUP_KEY}|{constants.METRIC_MEMORY_LIMIT_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.memory.node_reservation)))",
                    generate_metric,
                ),
                (
                    f"{constants.FILESYSTEM_GROUP_KEY}|{constants.METRIC_FILESYSTEM_AVAILABLE_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.filesystem.available)))",
                    generate_metric,
                ),
                (
                    f"{constants.FILESYSTEM_GROUP_KEY}|{constants.METRIC_FILESYSTEM_LIMIT_KEY}",
                    f"at(5m,align(5m,ts({metric_prefix}.node.filesystem.limit)))",
                    generate_metric,
                ),
                (
                    f"{constants.FILESYSTEM_GROUP_KEY}|{constants.METRIC_FILESYSTEM_USAGE_KEY}",
                    f"at(5m, align(5m,ts({metric_prefix}.node.filesystem.usage)) / align(5m,ts({metric_prefix}.node.filesystem.limit)) * 100 )",
                    generate_metric,
                ),
                (
                    constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY,
                    f"at(5m, sum(ts({metric_prefix}.node.pod.count), cluster, nodename))",
                    generate_metric,
                ),
                (
                    constants.PROPERTY_NODE_STATUS_KEY,
                    f"at(5m,ts({metric_prefix}.node.status.condition, condition=Ready))",
                    _generate_node_status_property,
                ),
            ]
        )

    nodes: dict[Key, Node] = {}
    async for attribute_key, timeseries, attribute_generating_function in get_timeseries(
        aoa_client, queries_and_metrics_keys
    ):
        try:
            for object_key, attribute in attribute_generating_function(
                attribute_key, timeseries, aoa_client.aria_apps_url, _get_key_from_tags
            ):
                # Add metric value to dict of metric values
                if object_key not in nodes:
                    new_node = Node(
                        object_key.get_identifier(constants.IDENTIFIER_CLUSTER_KEY),
                        object_key.get_identifier(constants.IDENTIFIER_NODE_KEY),
                        aoa_client.aria_apps_url,
                    )
                    new_node.with_property(
                        constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                    )
                    nodes[object_key] = new_node
                    logger.debug(f"Created new node '{object_key}'.")

                add_attribute(nodes[object_key], attribute)

        except Exception as e:
            logger.warning(f"Unable to add data to a Node: {e}")
            logger.exception(e)

    for node in nodes.values():
        for property in [
            constants.PROPERTY_INTERNAL_IP_KEY,
            constants.PROPERTY_PROVIDER_ID_KEY,
            constants.PROPERTY_NODE_ROLE_KEY,
        ]:
            result = node.get_last_property_value(property)
            if not result:
                node.with_property(property, "Unknown")
    logger.info(f"Collected {len(nodes)} total nodes")

    return nodes


def add_node_relationships(
    nodes: dict[Key, Node], clusters: dict[Key, Cluster]
) -> None:
    def get_cluster_key(node: Node) -> Optional[Key]:
        return cluster.get_key(
            node.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            node.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    add_relationships(clusters, nodes, logger, get_cluster_key)


# Helpers
def _get_key_from_tags(tags, aria_apps_url) -> Optional[Key]:
    node = tags.get("nodename")
    cluster = tags.get("cluster")
    return get_key(cluster, node, aria_apps_url)


def get_key(
    cluster: Optional[str], node: Optional[str], aria_apps_url: str
) -> Optional[Key]:
    if not node or not cluster:
        return None

    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.NODE_OBJECT_KIND,
        name=node,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_NODE_KEY, value=node),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )


def _generate_node_status_property(
    property_key: str,
    timeseries: dict,
    aria_apps_url: str,
    get_key_from_tags: FunctionType,
) -> Generator[tuple[Key, Property], None, None]:
    for ts in timeseries:
        tags = ts.get("tags", {})
        key = get_key_from_tags(tags, aria_apps_url)
        if ts["data"]:
            last_value_in_timeseries = ts["data"][-1][1]
            if key:
                yield key, Property(
                    property_key,
                    "Ready" if bool(last_value_in_timeseries) else "Not ready",
                )
        else:
            logger.warning(f"No data for '{property_key}' on {key}")
            yield key, Property(property_key, "Unknown")
