import json
import logging
import re
import time
from collections import defaultdict
from typing import List
from typing import Optional

import constants
import namespace
import node
import replica_set
from aoa_client import AOAClient
from aria.ops.adapter_instance import AdapterInstance
from aria.ops.definition.object_type import ObjectType
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from aria.ops.suite_api_client import SuiteApiClient
from error_handler import safe_call
from namespace import Namespace
from node import Node
from relationships import add_relationships
from replica_set import ReplicaSet

logger = logging.getLogger(__name__)


class Pod(Object):
    def __init__(self, cluster, namespace, pod, aria_apps_url):
        super().__init__(get_key(cluster, namespace, pod, aria_apps_url))

    @staticmethod
    def definition():
        pod_definition = ObjectType(constants.POD_OBJECT_KIND, "K8s Pod")

        # Identifiers
        pod_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster"
        )
        pod_definition.define_string_identifier(
            constants.IDENTIFIER_NAMESPACE_KEY, "Namespace"
        )
        pod_definition.define_string_identifier(constants.IDENTIFIER_POD_KEY, "Pod")
        pod_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        pod_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )
        pod_definition.define_string_property(
            constants.PROPERTY_NAMESPACE_NAME_KEY, "K8s Namespace"
        )
        pod_definition.define_string_property(
            constants.PROPERTY_NODE_NAME_KEY, "K8s Node"
        )
        pod_definition.define_string_property(
            constants.PROPERTY_REPLICA_SET_NAME_KEY, "K8s Replica Set"
        )
        pod_definition.define_string_property(
            constants.PROPERTY_POD_STATUS_KEY, "Status"
        )
        pod_definition.define_metric(
            constants.METRIC_LAST_UPDATE_TIME, "Last Update Time"
        )

        return pod_definition


@safe_call
async def get_pods(aoa_client: AOAClient, metric_prefixes: list[str]) -> dict[Key, Pod]:
    pods = {}
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        logger.info(f"Collecting pods for metric prefix '{metric_prefix}'")
        try:
            response = await aoa_client.run_query(
                f"at(5m,align(5m, ts({metric_prefix}.pod.status.phase)))"
            )
        except Exception as e:
            logger.exception(e)
            continue

        if "errorMessage" in response:
            logger.error(f"Error while fetching pod: {response['errorMessage']}")
            continue
        elif "warnings" in response:
            logger.warning(f"Warning while fetching pod: {response['warnings']}")

        if "timeseries" not in response:
            logger.warning(f"No pods returned for metric prefix '{metric_prefix}'")
            continue

        new_count = 0
        for pod_timeseries in response["timeseries"]:
            try:
                tags = pod_timeseries["tags"]

                pod = tags.pop("pod_name")
                cluster = tags.pop("cluster")
                namespace = tags.pop("namespace_name")
                node = tags.pop("nodename")
                phase = tags.get("phase", "Unknown")

                new_pod = Pod(cluster, namespace, pod, aoa_client.aria_apps_url)
                key = new_pod.get_key()

                if key in pods:
                    logger.warning(
                        f"Duplicate Pod key '{key}' was found during collection"
                    )
                    continue

                new_pod.with_property(
                    constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                )
                new_pod.with_property(constants.PROPERTY_NODE_NAME_KEY, node)
                new_pod.with_property(constants.PROPERTY_NAMESPACE_NAME_KEY, namespace)
                # Pods within a replicaset are named as the replicaset plus a 5-character
                # hash separated by a dash. The replicaset itself is named as the
                # deployment plus a 8-10 character hash. To avoid false positives as much
                # as possible, we look for both hashes and only fill in the replicaset
                # name if the pod name has the proper format.
                if re.search("-[0-9a-zA-Z]{8,10}-[0-9a-zA-Z]{5}$", key.name):
                    replicaset = re.sub("-[0-9a-zA-Z]{5}$", "", key.name)
                    new_pod.with_property(
                        constants.PROPERTY_REPLICA_SET_NAME_KEY, replicaset
                    )
                else:
                    new_pod.with_property(constants.PROPERTY_REPLICA_SET_NAME_KEY, "")
                new_pod.with_property(constants.PROPERTY_POD_STATUS_KEY, phase)
                new_pod.with_metric(constants.METRIC_LAST_UPDATE_TIME, time.time())
                pods[key] = new_pod
                logger.debug(f"Created new pod '{key}'.")
                new_count += 1

            except Exception as e:
                logger.warning(f"Unable to add data to a Pod due to: {e}")
                logger.exception(e)
        logger.info(f"Collected {new_count} pods for metric prefix '{metric_prefix}'")

    logger.info(f"Collected {len(pods)} total pods")
    return pods


def add_pod_relationships(
    pods: dict[Key, Pod],
    namespaces: dict[Key, Namespace],
    nodes: dict[Key, Node],
    replica_sets: dict[Key, ReplicaSet],
) -> None:
    def get_namespace_key(pod: Pod) -> Key:
        return namespace.get_key(
            pod.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            pod.get_identifier_value(constants.IDENTIFIER_NAMESPACE_KEY),
            pod.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    def get_node_key(pod: Pod) -> Key:
        return node.get_key(
            pod.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            pod.get_last_property_value(constants.PROPERTY_NODE_NAME_KEY),
            pod.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    def get_replicaset_key(pod: Pod) -> Optional[Key]:
        replica_set_name = pod.get_last_property_value(
            constants.PROPERTY_REPLICA_SET_NAME_KEY
        )
        # Pods do not need to be part of a replica set; if we can't find one return
        # 'None' so we don't attempt the relationship (and log the resulting failure)
        if replica_set_name:
            return replica_set.get_key(
                pod.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
                pod.get_identifier_value(constants.IDENTIFIER_NAMESPACE_KEY),
                replica_set_name,
                pod.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
            )
        else:
            return None

    add_relationships(namespaces, pods, logger, get_namespace_key)
    add_relationships(nodes, pods, logger, get_node_key)
    add_relationships(replica_sets, pods, logger, get_replicaset_key)

    for ns in namespaces.values():
        total = 0
        status_counts: dict[str, int] = defaultdict(lambda: 0)
        for child in ns.get_children():
            pod = pods.get(child)
            if pod:
                status = pod.get_last_property_value(constants.PROPERTY_POD_STATUS_KEY)
                status_counts[status] += 1
                total += 1
        ns.with_metric(constants.METRIC_TOTAL_NUMBER_OF_PODS_KEY, total)
        ns.with_metric(
            constants.METRIC_POD_STATE_KEY_RUNNING, status_counts.get("Running", 0)
        )
        ns.with_metric(
            constants.METRIC_POD_STATE_KEY_PENDING, status_counts.get("Pending", 0)
        )
        ns.with_metric(
            constants.METRIC_POD_STATE_KEY_SUCCEEDED, status_counts.get("Succeeded", 0)
        )
        ns.with_metric(
            constants.METRIC_POD_STATE_KEY_FAILED, status_counts.get("Failed", 0)
        )
        ns.with_metric(
            constants.METRIC_POD_STATE_KEY_UNKNOWN, status_counts.get("Unknown", 0)
        )


def get_key(
    cluster: Optional[str],
    namespace: Optional[str],
    pod: Optional[str],
    aria_apps_url: str,
) -> Optional[Key]:
    if not cluster or not namespace or not pod:
        return None
    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.POD_OBJECT_KIND,
        name=pod,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_NAMESPACE_KEY, value=namespace),
            Identifier(key=constants.IDENTIFIER_POD_KEY, value=pod),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )


def delete_not_existing_pods(
    client: SuiteApiClient, adapter_instance: AdapterInstance
) -> None:
    try:
        retention = adapter_instance.get_identifier_value(constants.DATA_RETENTION_TIME)
        if retention == constants.PLATFORM_DEFAULT or retention is None:
            logger.info(
                "Data retention time is set to platform default. Skipping Pod clean-up."
            )
            return
        adapter_instance_id = _get_adapter_instance_id(client, adapter_instance)
        if adapter_instance_id:
            logger.debug(
                f"Successfully obtained id='{adapter_instance_id}' for adapter "
                f"instance {adapter_instance.get_key()}"
            )
            not_existing_pods = _get_not_existing_resources(
                client, adapter_instance_id, adapter_instance
            )
            _delete_resources(client, not_existing_pods)
        else:
            logger.warning(
                f"Could not get 'id' for adapter instance {adapter_instance.get_key()}"
            )
    except Exception as e:
        logger.error(f"Unexpected error when cleaning up Pod resources: {e}")
        logger.exception(e)


def _get_adapter_instance_id(
    client: SuiteApiClient, adapter_instance: AdapterInstance
) -> Optional[str]:
    response = client.get("api/adapters?adapterKindKey=AriaApplicationsMP")
    if response.status_code < 300:
        target_url = adapter_instance.get_identifier_value(constants.ARIA_APPS_URL)
        for ai in json.loads(response.content).get("adapterInstancesInfoDto", []):
            for identifier in ai.get("resourceKey", {}).get("resourceIdentifiers", []):
                identifier_key = identifier.get("identifierType", {}).get("name", "")
                if identifier_key == constants.ARIA_APPS_URL:
                    if identifier.get("value", "") == target_url:
                        return ai.get("id")
    return None


def _get_not_existing_resources(
    client: SuiteApiClient, adapter_instance_id: str, adapter_instance: AdapterInstance
) -> List[str]:
    # Time window is in milliseconds since the epoch, but we're using seconds since the
    # epoch for this feature, so divide by 1000
    current_time = float(adapter_instance.collection_window["end_time"]) / 1000.0
    data_retention_time = (
        float(adapter_instance.get_identifier_value(constants.DATA_RETENTION_TIME))
        * 3600
    )
    cutoff_time = current_time - data_retention_time
    payload = {
        "adapterInstanceId": [adapter_instance_id],
        "resourceKind": [constants.POD_OBJECT_KIND],
        "resourceStatus": ["NO_DATA_RECEIVING", "NONE"],
        "statConditions": {
            "conjunctionOperator": "OR",
            "conditions": [
                {
                    "key": constants.METRIC_LAST_UPDATE_TIME,
                    "operator": "LT_EQ",
                    "doubleValue": cutoff_time,
                }
            ],
        },
    }
    try:
        logger.debug(
            f"Querying the list of Pods that have had a 'No Data Receiving' or 'None' "
            f"status since {cutoff_time} ({data_retention_time}s before present)."
        )
        resource_list = client.paged_post(
            "api/resources/query", json=payload, key="resourceList"
        ).get("resourceList", [])
        return [resource.get("identifier") for resource in resource_list]
    except Exception as e:
        logger.error(
            f"Could not retrieve list of Pods that have had a 'No Data Receiving' or "
            f"'None' status since {cutoff_time}."
        )
        logger.exception(e)
        return []


def _delete_resources(client: SuiteApiClient, uuids: List[str]) -> None:
    logger.info(f"{len(uuids)} Pods will be deleted.")
    logger.debug(f"Resource IDs of Pods: {repr(uuids)}")
    if not uuids:
        return
    payload = {"uuids": uuids}
    response = client.delete("api/resources/bulk", json=payload)
    if response.status_code < 300:
        logger.info(f"Successfully started batch delete job for {len(uuids)} Pods.")
    else:
        logger.error(f"Could not start batch delete job for {len(uuids)} Pods.")
