import logging
from typing import Optional

import constants
import deployment
from aenum import re
from aoa_client import AOAClient
from aria.ops.definition.object_type import ObjectType
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from deployment import Deployment
from error_handler import safe_call
from relationships import add_relationships

logger = logging.getLogger(__name__)


class ReplicaSet(Object):
    def __init__(self, cluster, namespace, replicaset, aria_apps_url):
        super().__init__(get_key(cluster, namespace, replicaset, aria_apps_url))

    @staticmethod
    def definition():
        replicaset_definition = ObjectType(
            constants.REPLICASET_OBJECT_KIND, "K8s ReplicaSet"
        )

        # Identifiers
        replicaset_definition.define_string_identifier(
            constants.IDENTIFIER_CLUSTER_KEY, "Cluster"
        )
        replicaset_definition.define_string_identifier(
            constants.IDENTIFIER_NAMESPACE_KEY, "Namespace"
        )
        replicaset_definition.define_string_identifier(
            constants.IDENTIFIER_REPLICASET_KEY, "ReplicaSet"
        )
        replicaset_definition.define_string_identifier(
            constants.IDENTIFIER_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        # Properties
        replicaset_definition.define_string_property(
            constants.PROPERTY_ARIA_APPS_URL_KEY,
            "VMware Aria Operations for Applications URL",
        )

        return replicaset_definition


@safe_call
async def get_replicasets(
    aoa_client: AOAClient, metric_prefixes
) -> dict[Key, ReplicaSet]:
    replicasets = {}
    for metric_prefix in metric_prefixes:
        if len(metric_prefix) < 1:
            continue
        logger.info(f"Collecting replicasets for metric prefix '{metric_prefix}'")
        try:
            response = await aoa_client.run_query(
                f"at(5m,rawsum(ts({metric_prefix}.replicaset.available_replicas), cluster, replicaset, namespace_name))",
            )
        except Exception as e:
            logger.exception(e)
            continue

        if "errorMessage" in response:
            logger.error(
                f"Error while fetching replicasets: {response['errorMessage']}"
            )
            continue
        elif "warnings" in response:
            logger.warning(
                f"Warning while fetching replicatsets: {response['warnings']}"
            )

        if "timeseries" not in response:
            logger.warning(
                f"No replicasets returned for metric prefix '{metric_prefix}'"
            )
            continue

        new_count = 0
        for replicaset_timeseries in response["timeseries"]:
            try:
                tags = replicaset_timeseries["tags"]

                cluster = tags.pop("cluster")
                namespace = tags.pop("namespace_name")
                replicaset = tags.pop("replicaset")

                new_replicaset = ReplicaSet(
                    cluster, namespace, replicaset, aoa_client.aria_apps_url
                )
                key = new_replicaset.get_key()

                if key in replicasets:
                    logger.warning(
                        f"Duplicate ReplicaSet key '{key}' was found during collection"
                    )
                    continue

                new_replicaset.with_property(
                    constants.PROPERTY_ARIA_APPS_URL_KEY, aoa_client.aria_apps_url
                )
                replicasets[key] = new_replicaset
                logger.debug(f"Created new replicaset '{key}'.")
                new_count += 1

            except Exception as e:
                logger.warning(f"Unable to add data to a ReplicaSet due to: {e}")
                logger.exception(e)
        logger.info(
            f"Collected {new_count} replicasets for metric prefix '{metric_prefix}'"
        )

    logger.info(f"Collected {len(replicasets)} total replicasets")
    return replicasets


def add_replicaset_relationships(
    replicasets: dict[Key, ReplicaSet], deployments: dict[Key, Deployment]
) -> None:
    def get_deployment_key(replicaset: ReplicaSet) -> Key:
        deployment_name = re.sub("-[0-9a-zA-Z]{8,10}$", "", replicaset.get_key().name)
        return deployment.get_key(
            replicaset.get_identifier_value(constants.IDENTIFIER_CLUSTER_KEY),
            replicaset.get_identifier_value(constants.IDENTIFIER_NAMESPACE_KEY),
            deployment_name,
            replicaset.get_identifier_value(constants.IDENTIFIER_ARIA_APPS_URL_KEY),
        )

    add_relationships(deployments, replicasets, logger, get_deployment_key)


def get_key(
    cluster: Optional[str],
    namespace: Optional[str],
    replicaset: Optional[str],
    aria_apps_url: str,
) -> Optional[Key]:
    if not replicaset or not namespace or not cluster:
        return None
    return Key(
        adapter_kind=constants.ADAPTER_KIND,
        object_kind=constants.REPLICASET_OBJECT_KIND,
        name=replicaset,
        identifiers=[
            Identifier(key=constants.IDENTIFIER_CLUSTER_KEY, value=cluster),
            Identifier(key=constants.IDENTIFIER_NAMESPACE_KEY, value=namespace),
            Identifier(key=constants.IDENTIFIER_REPLICASET_KEY, value=replicaset),
            Identifier(key=constants.IDENTIFIER_ARIA_APPS_URL_KEY, value=aria_apps_url),
        ],
    )
