import json
import logging

import constants
from aria.ops.object import Identifier
from aria.ops.object import Key
from aria.ops.object import Object
from aria.ops.suite_api_client import SuiteApiClient
from node import Node

logger = logging.getLogger(__name__)


class VirtualMachine(Object):
    def __init__(self, vm_name, identifiers) -> None:
        super().__init__(
            Key(
                adapter_kind=constants.VMWARE_ADAPTER_KIND,
                object_kind=constants.VMWARE_VM_KIND,
                name=vm_name,
                identifiers=identifiers,
            )
        )


def get_virtual_machines(
    client: SuiteApiClient, nodes: dict[Key, Node]
) -> list[VirtualMachine]:
    virtual_machines: dict = {}
    # build a query to fetch the virtual machines with the given node IP addresses
    for key, node in nodes.items():
        try:
            logger.debug(f"node key: {node.get_key()}:")
            provider_id = node.get_last_property_value(
                constants.PROPERTY_PROVIDER_ID_KEY
            )
            # We should not try to find a VR for nodes hosted on AWS
            if "aws" in provider_id:
                logger.debug(f"Skipping AWS node {key.name}")
                continue

            ip = node.get_last_property_value(constants.PROPERTY_INTERNAL_IP_KEY)
            query = {
                "adapterKind": [constants.VMWARE_ADAPTER_KIND],
                "resourceKind": [constants.VMWARE_VM_KIND],
                "propertyConditions": {
                    "conditions": [
                        {
                            "key": "net:4000|ip_address",
                            "operator": "EQ",
                            "stringValue": ip,
                        }
                    ]
                },
            }

            response = client.post(url="/api/resources/query", json=query)
            logger.debug(
                f"request {repr(query)} responded w/" + repr(response.status_code)
            )

            if response.status_code >= 300:
                logger.warning(
                    "skipping collection for virtual machine due to server response"
                )
                continue

            content = response.content.decode("utf-8")
            parsed = json.loads(content)

            if not len(parsed.get("resourceList", [])):
                logger.warning(f"No Virtual Machine found for node '{key.name}'")
                continue

            virtual_machine = parsed["resourceList"][0]
            # create the resource
            name = virtual_machine["resourceKey"]["name"]

            identifiers = []
            for identifier in virtual_machine["resourceKey"]["resourceIdentifiers"]:
                identifiers.append(
                    Identifier(
                        identifier["identifierType"]["name"],
                        identifier["value"],
                        identifier["identifierType"]["isPartOfUniqueness"],
                    )
                )

            new_virtual_machine = VirtualMachine(name, identifiers)
            key = new_virtual_machine.get_key()

            if key in virtual_machines:
                logger.warning(
                    f"Duplicate Virtual Machine with key {key} found while creating relationships"
                )
                new_virtual_machine = virtual_machines[key]
            else:
                virtual_machines[key] = new_virtual_machine

            new_virtual_machine.add_child(node)

        except Exception as e:
            logger.error(
                "Suite API collection failed while trying to fetch virtual machine "
                f"related to {key.name}"
            )
            logger.exception(e)

    return list(virtual_machines.values())
